///////////////////////////////////////////////////////////////////////////
////                                                                   ////
////                          XMLEngine.C                              ////
////                                                                   ////
//// Routines for reading and sending XML files.  These routines are   ////
//// designed to operate with minimal resources, for example:          ////
//// the parse routine doesn't need the whole XML file in memory at    ////
//// once, it can operate on the file at chunks at a time.             ////
////                                                                   ////
//// Please see XMLEngine.h for documentation.                         ////
////                                                                   ////
///////////////////////////////////////////////////////////////////////////
////                                                                   ////
//// VERSION HISTORY:                                                  ////
////                                                                   ////
//// Author Date        Comments                                       ////
//// -----  -------     ----------------------------------             ////
//// DSR    Aug-21-2006 Initial Release                                ////
//// DSR    Nov-02-2006 XMLSendDocument() has a new line on close tags.////
////                    XMLSendDocument() has tabs on open tags.       ////
//// TMH    Jun-22-2009 Added comments to some functions               ////
////                                                                   ////
///////////////////////////////////////////////////////////////////////////
////        (C) Copyright 1996, 2006 Custom Computer Services          ////
//// This source code may only be used by licensed users of the CCS C  ////
//// compiler.  This source code may only be distributed to other      ////
//// licensed users of the CCS C compiler.  No other use, reproduction ////
//// or distribution is permitted without written permission.          ////
//// Derivative programs created using this software in object code    ////
//// form are not restricted in any way.                               ////
///////////////////////////////////////////////////////////////////////////

#include <XMLParse.h>
#include <XMLEngine.h>

//#define debug_xmlsend   debug_printf
#define debug_xmlsend

//#define debug_xmlget   debug_printf
#define debug_xmlget(a,b,c,d,e,f,g,h,i,j,k)


#if (XML_PARSE_MAX_ELEMENT_VALUE_SIZE > XML_PARSE_MAX_ELEMENT_NAME_SIZE)
 #define XML_ENGINE_SCRATCH_SPACE   XML_PARSE_MAX_ELEMENT_VALUE_SIZE
#else
 #define XML_ENGINE_SCRATCH_SPACE   XML_PARSE_MAX_ELEMENT_NAME_SIZE
#endif

#ifndef XML_ENGINE_MAX_PARENTS
#define XML_ENGINE_MAX_PARENTS   5  //needs to be at least 1
#endif
      	
int16 *strlen16(char *s)
{
   char *sc;
   //int16 ret;

   for (sc = s; *sc != 0; sc++);
   
   return(sc - s);
//   ret = sc - s;
//   return(ret);
}

//d - destination string
//s - source string (null terminated)
//n - maximum space left in d, NOT COUNTING NULL TERMINATION
//returns - the number of chars read from s and saved to d, not including the
// null termination.
//this will always append a null termination at the end of d.
int16 strcpysafe(char *d, char *s, int16 n)
{
   char c;
   int16 i=0;

   //we pre-dec n to save 1 char space for null termination
   while (n-- && ((c=*s++)!=0))
   {
      *d++ = c;
      i++;
   }
   *d = 0;

   return(i);
}

//d - destination string
//s - source string (null terminated)
//n - maximum space left in d, counting the null termination
//returns - the number of chars read from s and saved to d, not including the
// null termination
/*
int16 strcpysafe(char *d, char *s, int16 n)
{
   char c;
   int16 i=0;

   //we pre-dec n to save 1 char space for null termination
   if (n)
   while (--n && ((c=*s++)!=0))
   {
      *d++ = c;
      i++;
   }
   *d = 0;

   return(i);
}
*/

static enum
{
   XML_GET_DOCUMENT_SM_START = 0,
   XML_GET_DOCUMENT_SM_CONTINUE = 1,
   XML_GET_DOCUMENT_SM_STOP = 2
} g_XMLGetDocumentSM;

//returns the XML_ESCAPE_CODE index for a parent that matches.
//returns -1 if not found or valid
static XML_ESCAPE_CODE XMLFindMatchingParent(char *newStr, XML_ESCAPE_CODE currParentRow)
{
   XML_ENGINE_ELEMENTS scratchXMLElement;
   XML_ESCAPE_CODE i;

   for (
         i=0;
         i < (sizeof(gc_XMLEngineElements)/sizeof(XML_ENGINE_ELEMENTS));
         i++
       )
   {
      scratchXMLElement = gc_XMLEngineElements[i];

      if (
            (scratchXMLElement.isParent) &&
            (!scratchXMLElement.isDisabled) &&
            (scratchXMLElement.parent == currParentRow) &&
            (strcmp(scratchXMLElement.elementName, newStr)==0)
         )
      {
         return(i);
      }
   }

   return(-1);
}

//returns TRUE if matched, FALSE if not found
//see XMLEngine.h for documentation
//TODO retEscapeCode is obsolete
int8 XMLFindMatchingEscapeElement
(
   char *elementName, XML_ESCAPE_CODE currParentRow,
   XML_ESCAPE_CODE *retEscapeCode, XML_ESCAPE_CODE *retFoundRow
)
{
   XML_ENGINE_ELEMENTS scratchXMLElement;
   XML_ESCAPE_CODE i;

   debug_xmlget("\r\n=== FIND %S %U ", elementName, currParentRow);

   for (
         i=0;
         i < (sizeof(gc_XMLEngineElements)/sizeof(XML_ENGINE_ELEMENTS));
         i++
       )
   {
      scratchXMLElement = gc_XMLEngineElements[i];

      if (
            (
               (scratchXMLElement.parent == currParentRow) ||
               (currParentRow == 0xFE)
            ) &&
            (!scratchXMLElement.isDisabled) &&
            (stricmp(scratchXMLElement.elementName, elementName)==0)
         )
      {
         if (retFoundRow)
            *retFoundRow = i;
         if (retEscapeCode)
            //*retEscapeCode = scratchXMLElement.escapeCode;
            *retEscapeCode = i;

         //debug_xmlget("@ row=%U escape=0x%X", i, scratchXMLElement.escapeCode);
         debug_xmlget("@ row=%U", i);

         return(TRUE);
      }
   }

   return(FALSE);
}

XML_ESCAPE_CODE * gp_XMLResponses;
int g_XMLMaxNumResponses;
int g_XMLNumResponses;

int XMLGetDocumentResponses(void)
{
   return(g_XMLNumResponses);
}

void XMLGetDocumentStart(XML_ESCAPE_CODE *responses, int responseMaxCount)
{
   gp_XMLResponses = responses;
   g_XMLMaxNumResponses = responseMaxCount;
   g_XMLNumResponses = 0;
   g_XMLGetDocumentSM = XML_GET_DOCUMENT_SM_START;
}

//in - the next chunk of the document (null terminated)
//
//always continues from the last position.  if this is a new document then
//use XMLGetDocumentStart() to initialize the state machine.
XML_ENGINE_EC XMLGetDocument(char *in)
{
   XML_PARSE_EC parseEC;
   static int8 openParentsIndex;
   static XML_ESCAPE_CODE openParentsRow[XML_ENGINE_MAX_PARENTS];
   static int8
      openParentsStr[XML_ENGINE_MAX_PARENTS][XML_PARSE_MAX_ELEMENT_NAME_SIZE];
   static int8 name[XML_PARSE_MAX_ELEMENT_NAME_SIZE];
   static int8 val[XML_PARSE_MAX_ELEMENT_VALUE_SIZE];
   static char *currParentStr, *newParentStr;
   XML_ESCAPE_CODE currParentRow, newParentRow, escapeCode, matchedRow;
   int8 ec;

   if (g_XMLGetDocumentSM == XML_GET_DOCUMENT_SM_START)
   {
      XMLStartParse(in);
      openParentsIndex = 0;      //there are currently no open parents
      currParentStr = NULL;
      newParentStr = &openParentsStr[0][0];
      openParentsStr[0][0]=0;
      name[0]=0;
      val[0]=0;
      g_XMLGetDocumentSM = XML_GET_DOCUMENT_SM_CONTINUE;
   }
   else
   {
      XMLContinueParse(in);
   }

   while (g_XMLGetDocumentSM == XML_GET_DOCUMENT_SM_CONTINUE)
   {
      parseEC = XMLGetElement(currParentStr, newParentStr, name, val);

      switch(parseEC)
      {
         case XML_EC_UNFINISHED:
            return(XML_ENGINE_EC_CONTINUE);

         case XML_EC_NEW_PARENT:
            if (openParentsIndex)
               currParentRow = openParentsRow[openParentsIndex-1];
            else
               currParentRow = -1;

            newParentRow = XMLFindMatchingParent(newParentStr, currParentRow);

            debug_xmlget("\r\n=== NEW PARENT %S %U %U",
                  newParentStr,
                  currParentRow,
                  newParentRow
               );

            //this new parent is good!
            if (newParentRow != -1)
            {
               openParentsRow[openParentsIndex] = newParentRow;
               currParentStr = &openParentsStr[openParentsIndex][0];
               openParentsIndex++;
               newParentStr = &openParentsStr[openParentsIndex][0];
            }
            else
            {
               return(XML_ENGINE_EC_UNKOWN_PARENT);
            }
            break;

         case XML_EC_CLOSE_PARENT:
            debug_xmlget("\r\n=== CLOSE PARENT %U", openParentsIndex);
            openParentsIndex--;
            newParentStr = &openParentsStr[openParentsIndex][0];
            if (openParentsIndex == 0)
            {  //root element just got closed
               currParentStr = NULL;
               g_XMLGetDocumentSM = XML_GET_DOCUMENT_SM_STOP;
            }
            else
            {
               currParentStr = &openParentsStr[openParentsIndex-1][0];
            }
            break;

         case XML_EC_OK:
            if (openParentsIndex)
               currParentRow = openParentsRow[openParentsIndex-1];
            else
               currParentRow = -1;   //for the rare event they have no child elements

            if (XMLFindMatchingEscapeElement(name, currParentRow, &escapeCode, &matchedRow))
            {
               if (strlen16(val))
               {
                  ec=XMLEngineWriteEscape(escapeCode, val);
                  if (ec)
                     return(ec);
               }

               if (g_XMLNumResponses < g_XMLMaxNumResponses)
               {
                  g_XMLNumResponses++;
                  *gp_XMLResponses++ = matchedRow;
               }

               if (currParentRow == -1)
               {
                  //he just asked for the root element
                  currParentStr = NULL;
                  g_XMLGetDocumentSM = XML_GET_DOCUMENT_SM_STOP;
               }
            }
            else
               return(XML_ENGINE_EC_UNKOWN_ELEMENT);
            break;

         default:
            return(parseEC);
      }
   }
   return(XML_ENGINE_EC_OK);
}

static enum
{
   XML_SEND_DOCUMENT_SM_START=0,
   XML_SEND_DOCUMENT_SM_NEXT_ELEMENT,
   XML_SEND_DOCUMENT_SM_NEXT_ELEMENT_CONTINUE,
   XML_CHECK_PARENTS,
   XML_OPEN_NEXT_PARENT,           
   XML_CLOSE_CURRENT_PARENT,
   XML_SEND_ELEMENT_OPEN,
   XML_SEND_ELEMENT_VALUE,
   XML_SEND_ELEMENT_CLOSE,
   XML_SEND_DOCUMENT_NEXT_CHILD,
   XML_SEND_DOCUMENT_NEXT_CHILD_CONTINUE,
   XML_SEND_OPEN_ELEMENT,
   XML_SEND_CLOSE_ELEMENT,
   XML_SEND_DOCUMENT_SM_SEND_SCRATCH,
   XML_SEND_DOCUMENT_SM_RETURN
} g_XMLSendDocumentSM;

XML_ESCAPE_CODE * gp_XMLSendElements;
int g_XMLSendElementsCount;
//int g_XMLSendElementsSecurity;

void XMLSendDocumentStart(XML_ESCAPE_CODE *elements, int elementsCount)
{
   g_XMLSendDocumentSM = XML_SEND_DOCUMENT_SM_START;
   gp_XMLSendElements = elements;
   g_XMLSendElementsCount = elementsCount;
   //g_XMLSendElementsSecurity = security;
}

// elementIndex - the element we want to find parents of
// pParents - an array of all currently opened parents (array is 
//             XML_ENGINE_MAX_PARENTS large).  pParents[0] is the first opened
//             parent, parents[1] is the sub-parent of pParents[0], etc.  If 
//             pParents[x]=-1 that means it is not currently open.
// Returns - if result is positive, that means new parents have been added to
//             *parents.  if result is 0, no new parents need to be opened.
//             if result is negative, that means you need to close that many
//             parents before continuing (parents will NOT be updated).  If 
//             result was negative, after you close the parents you need to 
//             run this routine again to see if you need to open any new 
//             parents.
static signed int FindParents(XML_ESCAPE_CODE elementIndex, XML_ESCAPE_CODE *pParents)
{
   int8 i=0;
   XML_ESCAPE_CODE j=0,k=0;
   signed int matchCount=0;
   XML_ESCAPE_CODE newParentsReverse[XML_ENGINE_MAX_PARENTS];
   XML_ESCAPE_CODE newParents[XML_ENGINE_MAX_PARENTS];
   XML_ESCAPE_CODE *p1,*p2;
   
   /*
 #if sizeof(XML_ESCAPE_CODE)==1
   UDPDebugPrintf(UDPDebug, "\nFindParents %X %X:%X:%X:%X:%X\f", 
 #else
   UDPDebugPrintf(UDPDebug, "\nFindParents %LX %LX:%LX:%LX:%LX:%LX\f", 
 #endif
         elementIndex,
         pParents[0],
         pParents[1],
         pParents[2],
         pParents[3],
         pParents[4]
      );*/
   
   while (i<XML_ENGINE_MAX_PARENTS)
   {
      elementIndex = gc_XMLEngineElements[elementIndex].parent;
      
      if (elementIndex != -1)
         newParentsReverse[i++]= elementIndex;
      else
         break;
   }
   
   //the above order has newParentsReverse in reverse order, swap it around
   memset(newParents, (XML_ESCAPE_CODE)-1, sizeof(newParents));
   while(i)
      newParents[j++]= newParentsReverse[--i];

   //find the result (see documentation above header)
   p1 = pParents;
   p2 = newParents;
   for (i=0; i<XML_ENGINE_MAX_PARENTS; i++)
   {
      j = *p1++;  //curr
      k = *p2++;  //new
      if (j==k)
         continue;
      else if ((j == -1)&&(k != -1))
      {
         matchCount = 1;
         break;
      }
      else
      {
         matchCount = -1;
         break;
      }
   }
   if (matchCount == 1)
   {
      matchCount = 0;
      do
      {
         matchCount++;
         i++;
         k = *p2++;
      } while ((i<XML_ENGINE_MAX_PARENTS)&&(k!=-1));
      memcpy(pParents, newParents, sizeof(newParents));
   }
   else if (matchCount != 0)
   {
      matchCount = 0;
      do
      {
         matchCount--;
         i++;
         j = *p1++;
      } while ((i<XML_ENGINE_MAX_PARENTS)&&(j!=-1));
   }
   
   /*
  #if sizeof(XML_ESCAPE_CODE)==1
   UDPDebugPrintf(UDPDebug,"\nFindParents Result count=%D %X:%X:%X:%X:%X\f"
  #else
   UDPDebugPrintf(UDPDebug,"\nFindParents Result count=%D %LX:%LX:%LX:%LX:%LX\f"
  #endif
         matchCount,
         pParents[0],
         pParents[1],
         pParents[2],
         pParents[3],
         pParents[4]
      );*/
   
   return(matchCount);
}

static int AddTabs(char *str, int8 tabs)
{
   int8 i=0;
   
   while(i++ < tabs)
      *str++ = '\t';
   
   return(tabs);
}

XML_ENGINE_EC XMLSendDocument(char *pResult, int16 maxSize)
{
   static int numParentsOpen;
   static signed int parentsToOpen;
   static XML_ESCAPE_CODE openParents[XML_ENGINE_MAX_PARENTS]={-1};
   static XML_ESCAPE_CODE childrenIndex=-1;
   static char scratchStr[XML_ENGINE_SCRATCH_SPACE + 10]; //10 = <, >, 5 tabs, \n and \r, null term
   static int nextSMAfterSendScratch, nextSMAfterElement;
   static char *sendPtr;
   static XML_ENGINE_ELEMENTS scratchXMLElement;
   int16 y;
   int1 newline=TRUE;
   char *debugPtr;
   XML_ESCAPE_CODE escape;
   
   debugPtr = pResult;

   maxSize--;  //reserve a space for null termination

   while(maxSize)
   {
      switch (g_XMLSendDocumentSM)
      {
         case XML_SEND_DOCUMENT_SM_START:
            numParentsOpen=0;
            memset(openParents, -1, sizeof(openParents));
            g_XMLSendDocumentSM = XML_SEND_DOCUMENT_SM_NEXT_ELEMENT_CONTINUE;
            break;
            
         case XML_SEND_DOCUMENT_SM_NEXT_ELEMENT:
            if (--g_XMLSendElementsCount==0)
            {
               parentsToOpen = -numParentsOpen;
               g_XMLSendDocumentSM = XML_CHECK_PARENTS;
               break;
            }
            else
            {
               gp_XMLSendElements++;
            }               
         case XML_SEND_DOCUMENT_SM_NEXT_ELEMENT_CONTINUE:
            childrenIndex = -1;
            parentsToOpen = FindParents(*gp_XMLSendElements, &openParents[0]);
            g_XMLSendDocumentSM = XML_CHECK_PARENTS;
            break;
         
         case XML_SEND_DOCUMENT_NEXT_CHILD:
            childrenIndex++;
         case XML_SEND_DOCUMENT_NEXT_CHILD_CONTINUE:
            if (childrenIndex < XML_ELEMENTS_COUNT)
            {
               scratchXMLElement = gc_XMLEngineElements[childrenIndex];
               escape = *gp_XMLSendElements;
               if (
                     !scratchXMLElement.isDisabled && 
                     !scratchXMLElement.isParent
                  )
               {
                  //go backwards until we find that this element is a child of our parent or hit the beginning
                  newline = scratchXMLElement.isHidden != 0;
                  while (
                           (scratchXMLElement.parent != -1) &&
                           (scratchXMLElement.parent != escape)
                        )
                  {
                     scratchXMLElement = gc_XMLEngineElements[scratchXMLElement.parent];
                     newline |= scratchXMLElement.isHidden != 0;
                  }
                  if ((scratchXMLElement.parent == escape) && !newline)
                  {
                     //this element is a child or sub-child of requested element
                     parentsToOpen = FindParents(childrenIndex, &openParents[0]);
                     g_XMLSendDocumentSM = XML_CHECK_PARENTS;
                  }
               }
            }
            else
            {
               childrenIndex = -1;
               g_XMLSendDocumentSM = XML_SEND_DOCUMENT_SM_NEXT_ELEMENT;
            }
            break;
            
         case XML_CHECK_PARENTS:
            if (parentsToOpen > 0)
               g_XMLSendDocumentSM = XML_OPEN_NEXT_PARENT;
            else if (parentsToOpen < 0)
               g_XMLSendDocumentSM = XML_CLOSE_CURRENT_PARENT;
            else
            {
               if (childrenIndex != -1)
               {
                  scratchXMLElement = gc_XMLEngineElements[childrenIndex];
                  nextSMAfterElement = XML_SEND_DOCUMENT_NEXT_CHILD;
                  g_XMLSendDocumentSM = XML_SEND_ELEMENT_OPEN;
               }
               else
               {
                  scratchXMLElement = gc_XMLEngineElements[*gp_XMLSendElements];
                  nextSMAfterElement = XML_SEND_DOCUMENT_SM_NEXT_ELEMENT;
                  if (!scratchXMLElement.isParent)
                     g_XMLSendDocumentSM = XML_SEND_ELEMENT_OPEN;
                  else
                     g_XMLSendDocumentSM = XML_SEND_DOCUMENT_NEXT_CHILD;
               }
            }
            break;

         case XML_OPEN_NEXT_PARENT:
            newline = TRUE;
            scratchXMLElement = gc_XMLEngineElements[openParents[numParentsOpen++]];
            parentsToOpen--;
            g_XMLSendDocumentSM = XML_SEND_OPEN_ELEMENT;
            nextSMAfterSendScratch = XML_CHECK_PARENTS;
            break;
            
         case XML_CLOSE_CURRENT_PARENT:
            numParentsOpen--;
            newline = TRUE;
            scratchXMLElement = gc_XMLEngineElements[openParents[numParentsOpen]];
            openParents[numParentsOpen]=-1;
            parentsToOpen++;  //logic is reversed because parentsToOpen is currently negative
            g_XMLSendDocumentSM = XML_SEND_CLOSE_ELEMENT;
            if (parentsToOpen < 0)
               nextSMAfterSendScratch = XML_CHECK_PARENTS;
            else
            {
               if (childrenIndex != -1)
               {
                  nextSMAfterSendScratch = XML_SEND_DOCUMENT_NEXT_CHILD_CONTINUE;
               }
               else
               {
                  if (g_XMLSendElementsCount)
                     nextSMAfterSendScratch = XML_SEND_DOCUMENT_SM_NEXT_ELEMENT_CONTINUE;
                  else
                     nextSMAfterSendScratch = XML_SEND_DOCUMENT_SM_RETURN;
               }
            }
            break;
         
         //next 3 states send a <element>value</element>
         case XML_SEND_ELEMENT_OPEN:
            newline = FALSE;
            g_XMLSendDocumentSM = XML_SEND_OPEN_ELEMENT;
            nextSMAfterSendScratch = XML_SEND_ELEMENT_VALUE;
            break;
            
         case XML_SEND_ELEMENT_VALUE:
            if (childrenIndex != -1)
               escape = childrenIndex;
            else
               escape = *gp_XMLSendElements;
            XMLEngineReadEscape(escape, &scratchStr[0], sizeof(scratchStr));
            sendPtr=scratchStr;
            g_XMLSendDocumentSM = XML_SEND_DOCUMENT_SM_SEND_SCRATCH;            
            nextSMAfterSendScratch = XML_SEND_ELEMENT_CLOSE;
            break;
            
         case XML_SEND_ELEMENT_CLOSE:
            newline = FALSE;
            g_XMLSendDocumentSM = XML_SEND_CLOSE_ELEMENT;
            nextSMAfterSendScratch = nextSMAfterElement;
            break;

         //send an open element string (ie <element>).  nextXMLSendDocumentSM
         //has already been preloaded with the next state machine.  set newline
         //to true if there should be a newline after the open element (which
         //you need to do if opening a parent, don't do it if you're about to
         //send a value)
         case XML_SEND_OPEN_ELEMENT:
            sendPtr = scratchStr;
            if (newline)
               sendPtr += AddTabs(sendPtr, numParentsOpen-1);
            else
               sendPtr += AddTabs(sendPtr, numParentsOpen);
            sprintf(sendPtr, "<%s>", scratchXMLElement.elementName);
            if (newline)
            {
               sendPtr += strlen16(sendPtr);
               sprintf(sendPtr, "\r\n");
            }
            sendPtr = scratchStr;
            g_XMLSendDocumentSM = XML_SEND_DOCUMENT_SM_SEND_SCRATCH;
            break;
         
         //send a close element string (ie </element>).  nextXMLSendDocumentSM
         //has already been preloaded with the next state machine.  set newline
         //to true if it tab over before sending the close element string.
         case XML_SEND_CLOSE_ELEMENT:
            sendPtr = scratchStr;
            if (newline)
               sendPtr += AddTabs(scratchStr, numParentsOpen);
            sprintf(sendPtr, "</%s>\r\n", scratchXMLElement.elementName);
            sendPtr = scratchStr;
            g_XMLSendDocumentSM = XML_SEND_DOCUMENT_SM_SEND_SCRATCH;
            break;

         //copy current contents of sendPtr into pResult.  if all contents of
         //sendPtr fit into pResult, then continue with onto the next state 
         //machine.  if we couldn't fit sendPtr into pResult then we need to 
         //continue this after the calling function has dealt with pResult.
         case XML_SEND_DOCUMENT_SM_SEND_SCRATCH:
            y = strcpysafe(pResult, sendPtr, maxSize);
            //debug_xmlsend(" ++ s%LX d%LX c%LU i%U j%U ", ptr, result, maxSize, i, j);
            pResult += y;
            maxSize -= y;
            //debug_xmlsend(" s%LX d%LX c%LU ++ ", ptr, result, maxSize);
            if (strlen16(sendPtr) == y)
            {
               if (nextSMAfterSendScratch == XML_SEND_DOCUMENT_SM_RETURN)
               {
                  return(XML_ENGINE_EC_OK);
               }
               else                  
                  g_XMLSendDocumentSM = nextSMAfterSendScratch;
            }
            else
            {
               sendPtr += y;  //continue from this point on next call
               return(XML_ENGINE_EC_CONTINUE);                 
            }               
            break;
      }
   }
   return(XML_ENGINE_EC_CONTINUE);
}

int XMLEngineFindEscapeCode(char *ptr, XML_ESCAPE_CODE *escapeCode, int8 showHidden)
{
   XML_ENGINE_ELEMENTS scratchXMLElement;
   int ret = FALSE;
   XML_ESCAPE_CODE i;
   
   for (i=0; i<=(sizeof(gc_XMLEngineElements)/sizeof(XML_ENGINE_ELEMENTS)); i++)
   {
      scratchXMLElement = gc_XMLEngineElements[i];
      //g_ATScratchElementEscapeCode = i;
      if ((showHidden || scratchXMLElement.isHidden == FALSE) && (scratchXMLElement.isDisabled == FALSE))
      {
         if (stricmp(scratchXMLElement.elementName, ptr)==0)
         {
            *escapeCode = i;
            ret = TRUE;
            break;
         }
      }
   }

   //fprintf(STREAM_DEBUG, "XML FIND '%s', FOUND=%U ESC=%U\r\n", ptr, ret, i);
   
   return(ret);
}

int1 XMLIsParent(XML_ESCAPE_CODE which)
{
   return(gc_XMLEngineElements[which].isParent);
}

//looks to see if 'child' is a child of 'parent'.  if it is a child, will
//return how many levels deep it is, or 0 if it is not a child.  if skipHidden
//is TRUE it will return false if any parent in the tree is hidden.
int XMLIsChild(XML_ESCAPE_CODE parent, XML_ESCAPE_CODE child)
{
   int level=0;
   XML_ESCAPE_CODE next;  
   XML_ENGINE_ELEMENTS scratchXMLElement;
  
   next = child;
  
   do
   {
      level++;
      scratchXMLElement = gc_XMLEngineElements[next];
      next = scratchXMLElement.parent;
   //} while ((next != -1) && (next != parent) && ((!skipHidden) || (skipHidden && !scratchXMLElement.isHidden)));
   } while ((next != -1) && (next != parent));

   if (scratchXMLElement.parent != parent)
      level = 0;
   
   return(level);
}

//set parent to -1 to continue previous
XML_ESCAPE_CODE XMLGetNextChild(XML_ESCAPE_CODE parent, XML_ESCAPE_CODE lastPosition=-1)
{
   static XML_ESCAPE_CODE currParent;
   static XML_ESCAPE_CODE currElement;
   XML_ENGINE_ELEMENTS scratchXMLElement;
   int1 parentIsHidden;
   
   if (parent != -1)
   {
      if (lastPosition == -1)
         currElement = 0;
      else
         currElement = ++lastPosition;
         
      currParent = parent;
   }
   else if (lastPosition == -1)
      currElement++;
   
   parentIsHidden = gc_XMLEngineElements[currElement].isHidden;
   
   while(currElement<XML_ELEMENTS_COUNT)
   {
      scratchXMLElement = gc_XMLEngineElements[currElement];
      if (
            !scratchXMLElement.isDisabled &&
            !scratchXMLElement.isParent &&
            XMLIsChild(currParent, currElement) &&
            (!scratchXMLElement.isHidden || parentIsHidden)
         )
      {
         return(currElement);
      }
      else
         currElement++;
   }
   
   return(-1); //not found
}



