///////////////////////////////////////////////////////////////////////////
////                                                                   ////
////                          XMLParse.c                               ////
////                                                                   ////
//// Routines for parsing an XML file.  These routines are designed to ////
//// operate with minimal resources, for example: the parse routine    ////
//// doesn't need the whole XML file in memory at once, it can operate ////
//// on the file at chunks at a time.                                  ////
////                                                                   ////
//// Please see XMLParse.h for documentation.                          ////
////                                                                   ////
///////////////////////////////////////////////////////////////////////////
////                                                                   ////
//// VERSION HISTORY:                                                  ////
////                                                                   ////
//// Author Date        Comments                                       ////
//// -----  -------     ----------------------------------             ////
//// DSR    Aug-21-2006 Initial Release                                ////
////                                                                   ////
///////////////////////////////////////////////////////////////////////////
////        (C) Copyright 1996, 2006 Custom Computer Services          ////
//// This source code may only be used by licensed users of the CCS C  ////
//// compiler.  This source code may only be distributed to other      ////
//// licensed users of the CCS C compiler.  No other use, reproduction ////
//// or distribution is permitted without written permission.          ////
//// Derivative programs created using this software in object code    ////
//// form are not restricted in any way.                               ////
///////////////////////////////////////////////////////////////////////////

#include <XMLParse.h>

//total space allocated for strings, this is the max number of chars (including
//null termination!) that has been allocated and that XMLGetElement() can write
//to.
#ifndef XML_PARSE_MAX_ELEMENT_NAME_SIZE
#define XML_PARSE_MAX_ELEMENT_NAME_SIZE  20
#endif

#ifndef XML_PARSE_MAX_ELEMENT_VALUE_SIZE
#define XML_PARSE_MAX_ELEMENT_VALUE_SIZE  80
#endif

#ifndef NULL
#define NULL   0
#endif

#include <string.h>

#define iswhitespace(c)   isamoung(c," \r\n\t")

char *gp_XMLLast;

static char XMLGetBuffer(void)
{
   char c;
   c = *gp_XMLLast++;
   return(c);
}

static void XMLUnGetBuffer(void)
{
   gp_XMLLast--;
}

enum {
   XMLGES_START = 0,
   XMLGES_GET_NAME_START = 1,
   XMLGES_GET_NAME = 2,
   XMLGES_GET_NAME_SKIP_ATTRIBUTE = 3,
   XMLGES_GET_NAME_ELEMENT = 4,
   XMLGES_GET_NAME_NONELEMENT = 5,
   XMLGES_GET_VAL = 6,
   XMLGES_GET_VAL_NOW = 7,
   XMLGES_GET_VAL_NEW_ELEMENT = 8,
   XMLGES_GET_VAL_CLOSE_ELEMENT = 9,
   XMLGES_GET_VAL_NONELEMENT = 10,
   XMLGES_GET_CLOSE = 11,
   XMLGES_CLOSE_PARENT = 12,
   XMLGES_GET_CLOSE_CONTINUE = 13,
   XMLGES_CLOSE_PARENT_CONTINUE = 14
} XMLGetElementState;

int1 XMLECIsSyntaxError(int ec)
{
   return(bit_test(ec,7));
}

void XMLContinueParse(char *in)
{
   gp_XMLLast = in;
}

void XMLStartParse(char *in)
{
   XMLGetElementState = XMLGES_GET_NAME_START;
   gp_XMLLast = in;
}

XML_PARSE_EC XMLGetRootElement(char *root)
{
   char ec;

/*
   char c;
   while ((c=XMLGetBuffer())!=0)
   {
      if (c=='<')
         break;
      else if (!iswhitespace(c))
         return(XML_EC_SYNTAX_ERROR_BAD_START);
   }
*/

   //XMLStartParse(in);

   ec = XMLGetElement(NULL, root, NULL, NULL);

   if (ec == XML_EC_NEW_PARENT)
      ec = XML_EC_OK;

   return(ec);
}

XML_PARSE_EC XMLGetElement(char *currParent, char *newParent, char *name, char *val)
{
   char c, *nameBegin;
   
   int16 j;

   static char prevState;
   static char lastc;
   static char *pCloseCompare;
   
   int16 maxVal=XML_PARSE_MAX_ELEMENT_VALUE_SIZE-1;
   int maxName=XML_PARSE_MAX_ELEMENT_NAME_SIZE-1;
   int maxParent=XML_PARSE_MAX_ELEMENT_NAME_SIZE-1;

   char ret;

   nameBegin = name;

   if (
         (XMLGetElementState != XMLGES_GET_NAME_START) &&
         (XMLGetElementState != XMLGES_GET_NAME)
      )
   {
      //we need to continue from previous location
      if (val)
      {
         j = strlen(val);
         maxVal -= j;
         val += j;        
      }
      if (name)
      {
         j = strlen(name);
         name += j;
         maxName -= j;
      }
      if (newParent)
      {
         j = strlen(newParent);
         maxParent -= j;
         newParent += j;
      }
   }

   while ( (c=XMLGetBuffer()) != 0 )
   {

      switch(XMLGetElementState)
      {
         case XMLGES_GET_NAME_START:            //after a non element, we need to get to the next element
            if (c == '<')
               XMLGetElementState = XMLGES_GET_NAME;
            else if (!iswhitespace(c))
               return(XML_EC_SYNTAX_ERROR_BAD_START);
            break;

         case XMLGES_GET_NAME:                  //we are on the first character of the element name
            if (val) *val=0;
            if (name) *name=0;
            if (newParent) *newParent=0;

         case XMLGES_GET_NAME_SKIP_ATTRIBUTE:   //skip any attributes associated with element
         case XMLGES_GET_NAME_NONELEMENT: //element started with a ! or ?, it's a non element (dcd maybe)
         case XMLGES_GET_NAME_ELEMENT:    //element started with an alpha-numeric, read and save to name (or parent)
            if (c == '>')
            {
               if (XMLGetElementState == XMLGES_GET_NAME_NONELEMENT)
                  XMLGetElementState = XMLGES_GET_NAME_START;
               else if (XMLGetElementState == XMLGES_GET_NAME_ELEMENT)
               {
                  if (lastc == '/')
                  {
                     XMLGetElementState = XMLGES_GET_NAME_START;
                     if (name)
                        return(XML_EC_OK);
                     else
                        return(XML_EC_CLOSE_PARENT);
                  }
                  else
                     XMLGetElementState = XMLGES_GET_VAL;
               }
               else
                  XMLGetElementState = XML_EC_SYNTAX_ERROR_BAD_CHAR;
            }
            else if (
                        (XMLGetElementState == XMLGES_GET_NAME_NONELEMENT) ||
                        (XMLGetElementState == XMLGES_GET_NAME_SKIP_ATTRIBUTE)
                    )
            {
               //skip the rest.  sorry for the cheap hack.
            }
            else if ((c=='?')||(c=='!'))
            {
               if (XMLGetElementState == XMLGES_GET_NAME)
                  XMLGetElementState = XMLGES_GET_NAME_NONELEMENT;
               else
                  return(XML_EC_SYNTAX_ERROR_BAD_CHAR);
            }
            else if (c=='/')
            {
               if (XMLGetElementState == XMLGES_GET_NAME)
                  XMLGetElementState = XMLGES_CLOSE_PARENT;
            }
            else if (isalnum(c))
            {
               if (XMLGetElementState == XMLGES_GET_NAME)
                  XMLGetElementState = XMLGES_GET_NAME_ELEMENT;
               if (XMLGetElementState == XMLGES_GET_NAME_ELEMENT)
               {
                  if (name)
                  {
                     if (maxName)
                     {
                        *name++ = c;
                        *name = 0;
                        maxName--;
                     }
                  }
                  else if (maxParent)  //we are getting root element
                  {
                     *newParent++ = c;
                     *newParent = 0;
                     maxParent--;
                  }
               }
            }
            else if (iswhitespace(c))
            {
               if (XMLGetElementState == XMLGES_GET_NAME_ELEMENT)
                  XMLGetElementState = XMLGES_GET_NAME_SKIP_ATTRIBUTE;
            }
            else
            {
               return(XML_EC_SYNTAX_ERROR_BAD_CHAR);
            }
            lastc = c;
            break;

         case XMLGES_GET_VAL:       //this will skip wait until the first non-ws char
         case XMLGES_GET_VAL_NOW:      //we have skipped all ws
         case XMLGES_GET_VAL_CLOSE_ELEMENT:    //we had gotten a < after some legal chars (a new element?)
         case XMLGES_GET_VAL_NEW_ELEMENT:    //the first non-ws char was a < (a new element?)
         case XMLGES_GET_VAL_NONELEMENT:  //looks like a comment is inside the value
            if (
                  (XMLGetElementState == XMLGES_GET_VAL_NEW_ELEMENT) ||
                  (XMLGetElementState == XMLGES_GET_VAL_CLOSE_ELEMENT)
               )
            {
               if (c=='/')
                  XMLGetElementState = XMLGES_GET_CLOSE;
               else if ((c=='!')||(c=='?'))
                  XMLGetElementState = XMLGES_GET_VAL_NONELEMENT;
               else if (XMLGetElementState == XMLGES_GET_VAL_NEW_ELEMENT)
               {
                  XMLUnGetBuffer();
                  if (name)
                     strcpy(newParent, nameBegin);
                  XMLGetElementState = XMLGES_GET_NAME;
                  return(XML_EC_NEW_PARENT);
               }
               else //(XMLGetElementState == XMLGES_GET_VAL_CLOSE_ELEMENT)
                  return(XML_EC_SYNTAX_ERROR_DOCUMENT);
            }
            else if (XMLGetElementState == XMLGES_GET_VAL_NONELEMENT)
            {
               if (c=='>')
                  XMLGetElementState = prevState;  //stop ignoring the comment
            }
            else if (c=='<')
            {
               prevState = XMLGetElementState;
               if (XMLGetElementState == XMLGES_GET_VAL)
                  XMLGetElementState = XMLGES_GET_VAL_NEW_ELEMENT;
               else
                  XMLGetElementState = XMLGES_GET_VAL_CLOSE_ELEMENT;
            }
            else if (isalnum(c) || c=='.' || (iswhitespace(c) && (XMLGetElementState==XMLGES_GET_VAL_NOW)))
            {
               XMLGetElementState = XMLGES_GET_VAL_NOW;
               if (val && maxVal)
               {
                  *val++ = c;
                  *val = 0;
                  maxVal--;
               }
            }
            else if (!iswhitespace(c) && (XMLGetElementState==XMLGES_GET_VAL))
            {
               return(XML_EC_SYNTAX_ERROR_BAD_CHAR);
            }
            break;

         case XMLGES_GET_CLOSE:  //read the close tag (we already got the /)
         case XMLGES_CLOSE_PARENT:
            if (XMLGetElementState == XMLGES_GET_CLOSE)
            {
               pCloseCompare = nameBegin;
               XMLGetElementState = XMLGES_GET_CLOSE_CONTINUE;
            }
            else
            {
               pCloseCompare = currParent;
               XMLGetElementState = XMLGES_CLOSE_PARENT_CONTINUE;
            }

         case XMLGES_GET_CLOSE_CONTINUE:
         case XMLGES_CLOSE_PARENT_CONTINUE:
            if (c=='>')
            {
               if (XMLGetElementState == XMLGES_GET_CLOSE_CONTINUE)
                  ret = XML_EC_OK;
               else
                  ret = XML_EC_CLOSE_PARENT;

               XMLGetElementState = XMLGES_GET_NAME_START;

               return(ret);
            }
            if (isalnum(c))
            {
               if (!pCloseCompare || (c != *pCloseCompare++))
               {
                  return(XML_EC_SYNTAX_ERROR_DOCUMENT);
               }
            }
            else
               return(XML_EC_SYNTAX_ERROR_BAD_CHAR);
            break;
      }
   }

   return(XML_EC_UNFINISHED);
}
