///////////////////////////////////////////////////////////////////////////
////                                                                   ////
////                          XMLParse.h                               ////
////                                                                   ////
//// Routines for parsing an XML file.  These routines are designed to ////
//// operate with minimal resources, for example: the parse routine    ////
//// doesn't need the whole XML file in memory at once, it can operate ////
//// on the file at chunks at a time.                                  ////
////                                                                   ////
///////////////////////////////////////////////////////////////////////////
////                                                                   ////
//// VERSION HISTORY:                                                  ////
////                                                                   ////
//// Author Date        Comments                                       ////
//// -----  -------     ----------------------------------             ////
//// DSR    Aug-21-2006 Initial Release                                ////
////                                                                   ////
///////////////////////////////////////////////////////////////////////////
////        (C) Copyright 1996, 2006 Custom Computer Services          ////
//// This source code may only be used by licensed users of the CCS C  ////
//// compiler.  This source code may only be distributed to other      ////
//// licensed users of the CCS C compiler.  No other use, reproduction ////
//// or distribution is permitted without written permission.          ////
//// Derivative programs created using this software in object code    ////
//// form are not restricted in any way.                               ////
///////////////////////////////////////////////////////////////////////////


#IFNDEF __PARSEXML_H__
#DEFINE __PARSEXML_H__

//////////////////////////////////////////////////////////////////////////////
//
// XML_EC
//
// Error codes for the XML Parser
//
// XML_EC_OK - Name and val have been updated for the next found element.
//             If <element></element> or <element/> then val will be
//             blank.
//
// XML_EC_UNFINISHED - We are in the middle of reading an XML element
//                     when *in was prematurely terminated.  You will
//                     have to call XMLGetElement again with the next block of
//                     input.
//
// XML_EC_NEW_PARENT - This element is a parent.  name and val will not be
//                     updated, newParent will hold the new element name.  If
//                     You used XMLGetRootElement() then this will be EC
//                     will be replaced with XML_EC_OK.
//
// XML_EC_CLOSE_PARENT - Our parent has just been closed (name and val
//                       will not be changed).
//
// XML_EC_SYNTAX_ERROR_WRONG_PARENT - A parent was closed that wasn't our
//             previously opened parent
//
// XML_EC_SYNTAX_ERROR_DOCUMENT - Reported if the XML is document type.  This
//             handler only supports data type XML.
//
// XML_EC_SYNTAX_ERROR_BAD_START - Reported if the XML document didn't start
//             with XML tags.
//
// XML_EC_SYNTAX_ERROR_BAD_CLOSE - An element was closed that wasn't the
//             previously opened element.
//
// XML_EC_SYNTAX_ERROR_BAD_CHAR - A bad character was found in the element
//
//////////////////////////////////////////////////////////////////////////////

typedef enum
{
   XML_EC_OK = 0,
   XML_EC_UNFINISHED = 1,
   XML_EC_NEW_PARENT = 2,
   XML_EC_CLOSE_PARENT = 3,
   XML_EC_SYNTAX_ERROR_WRONG_PARENT = 0x80,
   XML_EC_SYNTAX_ERROR_DOCUMENT = 0x81,
   XML_EC_SYNTAX_ERROR_BAD_START = 0x82,
   XML_EC_SYNTAX_ERROR_BAD_CLOSE = 0x83,
   XML_EC_SYNTAX_ERROR_BAD_CHAR = 0x84
} XML_PARSE_EC;

//////////////////////////////////////////////////////////////////////////////
//
// XMLStartParse(in)
//
// Tells XMLGetElement() where to start parsing data from.  Resets the
// Get Element state machine.  After XMLGetElement() returns XML_EC_UNFINISHED,
// you will need to use XMLContinueParse() to tell it where the next buffer to
// read is.
//
//////////////////////////////////////////////////////////////////////////////
void XMLStartParse(char *in);

//////////////////////////////////////////////////////////////////////////////
//
// XMLContinueParse(in)
//
// Tells XMLGetElement() where to continue parsing data from.  After
// XMLGetElement() returns XML_EC_UNFINISHED, you will need to use
// XMLContinueParse() to tell it where the next buffer to read is.
//
//////////////////////////////////////////////////////////////////////////////
void XMLContinueParse(char *in);

//////////////////////////////////////////////////////////////////////////////
//
// XMLGetRootElement(root)
//
// Uses XMLGetElement() to only get the Root Element.  Returns XML_EC_OK
// if a root element is found.  If it returns XML_EC_CLOSE_PARENT, that means
// the root element was closed instantly with shorthand notation.
// Else, see XML_EC for other possible errors.
//
//////////////////////////////////////////////////////////////////////////////
XML_PARSE_EC XMLGetRootElement(char *root);

//////////////////////////////////////////////////////////////////////////////
//
// XMLGetElement(currParent, newParent, name, val)
//
// Reads a string in memory and returns the first found element name/value.
//
// Inputs:
//   currParent - our current parent.  Set to NULL if you are looking for the
//            root element.
//   newParent - if a new parent has been started, it will be saved here and
//             it will return with XML_EC_NEW_PARENT (name and val will not
//             be updated)
//   name - Where the next found element name will be saved.
//          This can be set to NULL if you are looking for the root element,
//          but this is optional (if it is NULL it will use newParent for
//          scratch space while saving an element name)
//   value - Where the value for this element will be saved, if it exists.
//           If null pointer, don't save value.
//
// Returns:
//   A result code, XML_EC
//
// Limitations:
//    *  It only handles data xml, not document xml.
//       Therefore, this is illegal: <BODY><HTML><B>HELLO</B>THERE</HTML></BODY>
//       This is legal: <BODY><HTML>HELLO THERE</HTML></BODY>
//
//////////////////////////////////////////////////////////////////////////////
XML_PARSE_EC XMLGetElement(char *currParent, char *newParent, char *name, char *val);

#endif
