///////////////////////////////////////////////////////////////////////////
////                                                                   ////
////                          EX_XML.c                                 ////
////                                                                   ////
//// An Example showing the use of XMLEngine drivers. The program has  ////
//// an XML document containing a few elements. The entire XML document////
//// never has to be on the PIC at once, saving memory.                ////
////                                                                   ////
//// To use, connect the RS232. Follow the prompts on the screen. It is////
//// important that the chunk of XML that you send through (via the    ////
//// 's' command) never exceeds the amount of characters in the buffer ////
//// string. To see the program in action, you can send the following  ////
//// snippets of code.                                                 ////
//// You can read them back to verify correct behavior.                ////
////                                                                   ////
//// You can send the following via serial to change one element:      ////
//// 		<XML1000>                                                     ////                
////		   <CFG>                                                      ////              
////		      <VERSION>8</VERSION>                                    ////                                
////		   </CFG>                                                     ////               
////		</XML1000>                                                    //// 
////                                                                   ////
//// The XMLEngine is written in such a way to allow chunks of an XML  ////
//// file at a time instead of all at once. Send the following first:  ////
//// <XML1000>                                                         ////
////    <CFG>                                                          ////
////       <VERSION>10</VERSION>                                       ////
////    </CFG>                                                         ////
////    <PINS>                                                         ////
////       <PINA5>0</PINA5>                                            ////
////                                                                   ////
//// Then when prompted to continue the file, send this:               ////
////   <PINB4>0</PINB4>                                                ////
////     <PINB5>0</PINB5>                                              ////
////  </PINS>                                                          ////
//// </XML1000>                                                        ////
////                                                                   ////
//// When sending multiple files, the file size should be as large as  ////
//// the buffer (no larger) until the last file is reached. This       ////
//// ensures proper transfer of contents.                              ////
////																						  ////
//// The following MUST be provided by the user for the XML to work    ////
//// properly:                                                         ////
//// 	XML_ESCAPE_CODE																  ////
////  gc_XMLEngineElements[]														  ////
////  XMLEngineReadEscape()                                            ////
////  XMLEngineWriteEscape()														  ////
////                                                                   ////
///////////////////////////////////////////////////////////////////////////
////                                                                   ////
//// VERSION HISTORY:                                                  ////
////                                                                   ////
//// Author Date        Comments                                       ////
//// -----  -------     ----------------------------------             ////
//// TMH    Jun-22-2009 Completed example                              ////
////                                                                   ////
///////////////////////////////////////////////////////////////////////////
////        (C) Copyright 1996, 2009 Custom Computer Services          ////
//// This source code may only be used by licensed users of the CCS C  ////
//// compiler.  This source code may only be distributed to other      ////
//// licensed users of the CCS C compiler.  No other use, reproduction ////
//// or distribution is permitted without written permission.          ////
//// Derivative programs created using this software in object code    ////
//// form are not restricted in any way.                               ////
///////////////////////////////////////////////////////////////////////////

#include <18F4520.H>
#device ICD=TRUE
#fuses HS,NOWDT
#use delay(clock=20000000)
#use rs232(baud=9600, xmit=PIN_C6, rcv=PIN_C7)

//Maximum size for an element name
#define XML_PARSE_MAX_ELEMENT_NAME_SIZE  20

//Maximum size for an elements value
#define XML_PARSE_MAX_ELEMENT_VALUE_SIZE  80

//this enum MUST be in the EXACT order as gc_XMLEngineElements!
typedef enum
{
   XML_ELEMENT_XML1000=0,
   XML_ELEMENT_CFG,
   XML_ELEMENT_VERSION,
   XML_ELEMENT_ADC,
   XML_ELEMENT_AN0,
   XML_ELEMENT_AN1,
   XML_ELEMENT_PINS,
   XML_ELEMENT_PINA5,
   XML_ELEMENT_PINB4,
   XML_ELEMENT_PINB5,
} XML_ESCAPE_CODE;


//Must be copied directly between XML_ESCAPE_CODE and gc_XMLEngineElements[]
#define __XML_ENGINE_ELEMENTS__
typedef struct
{
   char elementName[XML_PARSE_MAX_ELEMENT_NAME_SIZE];
   XML_ESCAPE_CODE parent;          //set to -1 if root element (no parent)
   int isParent:1;     //set to TRUE if this is a parent (has no values, only children)
   int isHidden:1;        //if TRUE, hide this element when you read the entire parent
   int isDisabled:1;      //if TRUE, this element is disabled (you can't use read or write it).
                        //this is very similar to isHidden, provided so you can easily
                        //alter your XML_ENGINE_ELEMENTS array to conditionally compile
                        //specific tags.
} XML_ENGINE_ELEMENTS;

//Elements in the XML file
//{STRING, PARENT, IS PARENT, IS HIDDEN, IS DISABLED}
const XML_ENGINE_ELEMENTS gc_XMLEngineElements[]={
   {"XML1000", -1, TRUE, FALSE, FALSE}, 
   	{"CFG", XML_ELEMENT_XML1000, TRUE, FALSE, FALSE},
   		{"VERSION", XML_ELEMENT_CFG, FALSE, FALSE, FALSE},
      {"ADC", XML_ELEMENT_XML1000, TRUE, FALSE, FALSE},                         
         {"AN0", XML_ELEMENT_ADC, FALSE, FALSE, FALSE},
         {"AN1", XML_ELEMENT_ADC, FALSE, FALSE, FALSE},
      {"PINS", XML_ELEMENT_XML1000, TRUE, FALSE, FALSE},
   		{"PINA5", XML_ELEMENT_PINS, FALSE, FALSE, FALSE},
   		{"PINB4", XML_ELEMENT_PINS, FALSE, FALSE, FALSE},
   		{"PINB5", XML_ELEMENT_PINS, FALSE, FALSE, FALSE},
};

#include <XMLParse.c>
#include <XMLEngine.c>

//for the itoa and atoi functions
#include <stdlib.h>

//location where the VERSION number is stored in EEPROM
#define VERSION_EE_ADDR 0x0

//size of the buffer to hold the XML data
#define bufferSize 100

#word PORTA = getenv("sfr:PORTA")
#word PORTB = getenv("sfr:PORTB")

//Error code used by main
typedef enum{
	FILE_RETRIEVE_SUCCESS,
	INCORRECT_XML_FILE,
	FILE_TIMEOUT
}FILE_RETRIEVE_ERROR;

int1 timeoutFlag;

//Simulating a sent file, this is the default XML file on reset
char myXML[] = 
"<XML1000>
   <CFG>
      <VERSION>1</VERSION>
   </CFG>
   <PINS>
      <PINA5>1</PINA5>
      <PINB4>0</PINB4>
      <PINB5>1</PINB5>
   </PINS>
</XML1000>";

//size of the buffer to send from the XML document out/in to the serial port
//can differ to accomodate the memory needs of the program.
char buffer[bufferSize];

int8 elements[]={XML_ELEMENT_XML1000};
int8 responses[20];


//retrieves sent files via the serial port and calls getDocument to save the XML
//document to their respective locations
FILE_RETRIEVE_ERROR retrieveXML(){
	
	FILE_RETRIEVE_ERROR fec = FILE_RETRIEVE_SUCCESS;
	XML_ENGINE_EC ec;
	int32 timeout;
	int8 i;
	int1 firstParse = 1;
	
	
	do{
		timeoutFlag = 0;
		timeout = 0;
		i = 0;
		
		//waiting for first character, 10 second timeout                                
		while((!kbhit()) && (++timeout < 1000000)){
			delay_us(10);
		}
		
		if(!kbhit()){
			timeoutFlag = 1;
			fec = FILE_TIMEOUT;
		}else{
			while(!timeoutFlag){
			  	timeout = 0;
				while((++timeout < 50000) && (!kbhit())){
					delay_us(10);
			   } 
			   if(timeout > 49999){
					timeoutFlag = 1;   
				}else if(i < bufferSize){
					buffer[i] = getc();
					i++;
				}
			}
			
			//only call getdocumentstart once
			if(firstParse){
			    	XMLGetDocumentStart(responses, sizeof(responses));
			    	firstParse = 0;
			}
			
		   ec = XMLGetDocument(buffer);
			
			//checking if the file parsed correctly
			if(ec == XML_ENGINE_EC_CONTINUE){
				 printf("Input next part of XML file\r\n");
				 timeoutFlag = 0;
			}else if(ec != XML_ENGINE_EC_OK){
				 fec = INCORRECT_XML_FILE;     
			}
		}
	//continue if file is incomplete and not timed out
	}while((ec == XML_ENGINE_EC_CONTINUE) && (!timeoutFlag));
	return fec;
}


void main(void){
	char command;
	
	//Error codes
	XML_ENGINE_EC ec;
	FILE_RETRIEVE_ERROR fec;

	setup_adc(ADC_CLOCK_INTERNAL);
	setup_adc_ports(AN0_TO_AN1);
	
	//Getting myXML as a default document
	XMLGetDocumentStart(responses, sizeof(responses));
   XMLGetDocument(myXML);
	
	while(TRUE){
	
		  printf("\r\n(s)end or (r)etrieve an XML file?\r\n");
		  do{
		      command = toupper(getc());
		  }while((command != 'R') && (command != 'S'));
		  
		  //Send current document THROUGH RS232
		  if(command == 'R'){
		  		printf("XML SENDING\r\n\n");
	
				   XMLSendDocumentStart(&elements[0], sizeof(elements));
				   
				   do {
				      ec=XMLSendDocument(buffer, sizeof(buffer));
				      printf("%s",buffer);
				   } while (ec==XML_ENGINE_EC_CONTINUE);
				   
				   
				printf("\r\n\nDONE");  
				
		  //receive XML file(s) FROM RS232
		  }else if(command == 'S'){
		  
		  		printf("Send XML File\r\n");
		  		
		  		fec = retrieveXML();
		  		
		  		//checkin error code
		  		if(fec == FILE_RETRIEVE_SUCCESS){
		  		    printf("File Successfully Read\r\n");      
		  		}else if(fec == INCORRECT_XML_FILE){
		  		    printf("Incorrect XML File\r\n");      
		  		}else{
		  			 printf("Timed out\r\n");
		  		}
		  }
	  }
	
}


//Required callback function. Handles each escape code. See XMLEngine.h for 
//documentation.
int32 XMLEngineReadEscape(XML_ESCAPE_CODE escapeCode, char *ret, int16 maxSize){
	int32 retVal = 0;
	
	//XML_ELEMENT_XML1000, XML_ELEMENT_CFG, XML_ELEMENT_ADC, and XML_ELEMENT_PINS
	//are unnecessary   because they are parents and don't have values.                  

	switch(escapeCode){
		case XML_ELEMENT_VERSION:
			retVal = read_eeprom(VERSION_EE_ADDR);
			itoa(retVal,10, ret);   
		break;
		
   	case XML_ELEMENT_AN0:
   		set_adc_channel(0);
   		delay_us(10);
   		retVal = read_adc();
   		itoa(retVal,16, ret);
		break;
		
   	case XML_ELEMENT_AN1:
   		set_adc_channel(1);
   		delay_us(10);
   		retVal = read_adc();
   		itoa(retVal,16, ret);
		break;
		
   	case XML_ELEMENT_PINA5:
   		retVal = ((PORTA & 0x20) >> 5);
   		itoa(retVal,16, ret);
		break;
		
   	case XML_ELEMENT_PINB4:
   		retVal = ((PORTB & 0x10) >> 4);
   		itoa(retVal,2, ret);
		break;
		
   	case XML_ELEMENT_PINB5:
   		retVal = ((PORTB & 0x20) >> 5);
   		itoa(retVal,2, ret);
		break;
		
   	default:
	
	}
	
	return retVal;
}

//Required callback function. Handles each escape code. See XMLEngine.h for 
//documentation.
XML_ENGINE_EC XMLEngineWriteEscape(XML_ESCAPE_CODE escapeCode, char *str){

	XML_ENGINE_EC ec = XML_ENGINE_EC_OK;
	int16 value;
	
	value = atoi(str);

	switch(escapeCode){
		case XML_ELEMENT_VERSION:
			write_eeprom(VERSION_EE_ADDR, (int8)value);   
		break;
		
   	case XML_ELEMENT_AN0:
   		//PINA0 is a read-only ADC value
   		ec = XML_ENGINE_EC_BAD_DATA_RO;
		break;
		  
   	case XML_ELEMENT_AN1:
   		//PINA1 is a read-only ADC value
   		ec = XML_ENGINE_EC_BAD_DATA_RO;
		break;
		
   	case XML_ELEMENT_PINA5:
   		output_bit(PIN_A5, bit_test(value, 0));
		break;
		
   	case XML_ELEMENT_PINB4:
   		output_bit(PIN_B4, bit_test(value, 0));
		break;
		
   	case XML_ELEMENT_PINB5:
   		output_bit(PIN_B5, bit_test(value, 0));
		break;
	
	}

	return ec;
}

